
-- The card trick example, following the Huet more closely.

-- Identity ---------------------------------------------------------------

infix 5 ==

data (==) {A:Set}(a:A) : A -> Set where
  refl : a == a

cong (A,B:Set)(x,y:A) : (f : A -> B) -> x == y -> f x == f y
cong f refl = refl

-- Natural numbers --------------------------------------------------------

data Nat : Set where
  zero : Nat
  suc  : Nat -> Nat

infixl 10 +
infixl 12 *

(+) : Nat -> Nat -> Nat
0     + m = m
suc n + m = suc (n + m)

plus0 : (n : Nat) -> n + 0 == n
plus0 0 = refl
plus0 (suc n) = cong suc (plus0 n)

plusS : (n, m : Nat) -> n + suc m == suc (n + m)
plusS 0 m = refl
plusS (suc n) m = cong suc (plusS n m)

rewrite plus0; plusS

(*) : Nat -> Nat -> Nat
0     * m = 0
suc n * m = m + n * m

-- Parity -----------------------------------------------------------------

data Parity : Nat -> Set where
  even : (k:Nat) -> Parity (2 * k)
  odd  : (k:Nat) -> Parity (2 * k + 1)

parity : (n:Nat) -> Parity n
parity 0	= even 0
parity (suc n) with parity n =>
       (suc (2 * k))	 (even k) = odd k
       (suc (2 * k + 1)) (odd k)  = even (suc k)

-- Lists ------------------------------------------------------------------

infixr 15 ::

data List (A : Set) : Nat -> Set where
  nil	       : List A 0
  (::) (n:Nat) : A -> List A n -> List A (suc n)

-- Concatenation ----------------------------------------------------------

infixl 15 ++

(++) (A:Set)(n,m:Nat) : List A n -> List A m -> List A (n + m)
[]	++ ys = ys
(x::xs) ++ ys = x :: (xs ++ ys)

-- data Append {A:Set}{n,m:Parity} : List A n -> List A m -> List A (n + m) -> Set where
--   appNil (ys:List A)		 : Append [] ys ys
--   appCons (x:A)(xs,ys,zs:List A) : Append xs ys zs -> Append (x::xs) ys (x::zs)
-- 
-- catAppend (A:Set)(xs,ys,zs:List A) : Append xs ys zs -> xs ++ ys == zs
-- catAppend appNil	= refl
-- catAppend (appCons app) = cong ((::) _) (catAppend app)
-- 
-- assocCat (A:Set) : (xs,ys,zs:List A) -> xs ++ (ys ++ zs) == (xs ++ ys) ++ zs
-- assocCat [] ys zs	= refl
-- assocCat (x::xs) ys zs	= cong ((::) x) (assocCat xs ys zs)

-- Singleton --------------------------------------------------------------

-- singleton (A:Set) : A -> List A odd
-- singleton x = [x]

{-
rotate (A:Set) : List A -> List A
rotate []      = []
rotate (x::xs) = xs ++ [x]

data Bit : Set where
  O : Bit
  I : Bit

neg : Bit -> Bit
neg O = I
neg I = O

negNegId : (x : Bit) -> neg (neg x) == x
negNegId O = refl
negNegId I = refl

rewrite negNegId

data Alternating : Bit -> List Bit -> Set where
  altCns (x : Bit)(xs : List Bit) : Alternating (neg x) xs -> Alternating x (x::xs)
  altNil (x:Bit)		  : Alternating x []

data Paired : List Bit -> Set where
  pairCons (x : Bit)(xs : List Bit) : Paired xs -> Paired (x :: neg x :: xs)
  pairUnit (x : Bit)		    : Paired [x]
  pairNil			    : Paired []
-}
