-- The state monad transformer
-- Author :: Ulf Norell
-- Created:: 2004-12-16

-- # import Monad
-- # import Monad.Trans
-- # import Prelude.Proofs
-- # import EqualityReasoning

-- The state monad transformer --------------------------------------------

StateT : Set -> (Set -> Set) -> Set -> Set
StateT S M A = S -> M { A ; S }

monadTransStateT : (S : Set) -> MonadTransformer (StateT S)
monadTransStateT S =
    ( (monad, lift), proofs )
  where
    lift : (M : Set -> Set) -> MonadOps M -> (A : Set) -> M A -> StateT S M A
    lift M monadM A m = \s -> liftM (\x -> (x,s)) m
      where
        liftM = liftM M monadM A { A ; S }

    return : (M : Set -> Set) -> MonadOps M -> (A : Set) -> A -> StateT S M A
    return M monadM A x = lift' (ret x)
      where
        lift' = lift M monadM A
        ret   = monadM.return A

    bind : (M : Set -> Set) -> MonadOps M -> (A, B : Set) ->
            StateT S M A -> (A -> StateT S M B) -> StateT S M B
    bind M monadM A B m k = \s -> m s >>= \p -> k p.1 p.2
      where
        (>>=) = monadM.bind { A ; S } { B ; S }

    monad : (M : Set -> Set) -> MonadOps M -> MonadOps (StateT S M)
    monad M monadM = struct {return=return M monadM; bind = bind M monadM}

    monadLaws : (M : Set -> Set) -> (monadM : Monad M) ->
                  MonadLaws (StateT S M)  (return M monadM.ops)
                                          (bind M monadM.ops)
    monadLaws M monadM = struct
      law1 = -- The first monad law
        \A m -> fol-plugin( monadM.proofs.law1 {A;S}
                          , monadM.proofs.law2 A {A;S}
                          , propUnliftedPairs_ A S
                          , extensionality
                          )
      law2 =  -- The second monad law
        \A B x f -> fol-plugin( monadM.proofs.law2 A {A;S}
                              , monadM.proofs.law2 {A;S} {B;S}
                              , extensionality
                              , __fol_pair_elimination
                              )
      law3 = -- The third monad law
        -- Note:: Proof doesn't go through if we remove the m g h arguments.
        \A B C m g h -> ?
                      {-fol-plugin( monadM.proofs.law3 {A;S} {B;S} {C;S}
                                  , extensionality
                                  )-}

    proofs : MonadTransformerLaws (StateT S) monad lift
    proofs =
      ( monadLaws
      , \M monadM A x -> fol-plugin( extensionality )
      , \M monadM A B m k ->

          let bindAB    = monadM.ops.bind A B
              bindBBS   = monadM.ops.bind B {B;S}
              bindABS   = monadM.ops.bind A {B;S}
              bindAAS   = monadM.ops.bind A {A;S}
              bindASBS  = monadM.ops.bind {A;S} {B;S}
              returnBS  = monadM.ops.return {B;S}
              returnAS  = monadM.ops.return {A;S}
              (===)     = eqStep (S -> M {B;S})
          in

          extractProof (S -> M {B;S})
          ( beginEqProof (S -> M {B;S})

              (\s -> (m `bindAB` k) `bindBBS` \x -> returnBS (x,s))

                ===

              (\s -> m `bindABS` \x -> k x `bindBBS` \y -> returnBS (y,s))

                    `by`  fol-plugin( monadM.proofs.law3 A B {B;S}, extensionality )

                ===

              (\s -> m `bindABS` \x -> returnAS (x,s) `bindASBS`
                      \p -> k p.1 `bindBBS` \y -> returnBS (y,p.2)
              )
                    `by` ?
                        {-fol-plugin( monadM.proofs.law2 {A;S} {B;S}
                                    , extensionality
                                    , __fol_pair_elimination
                                    )-}

                ===

              (\s -> (m `bindAAS` \x -> returnAS (x,s)) `bindASBS`
                      \p -> k p.1 `bindBBS` \y -> returnBS (y,p.2)
              )
                  `by` ?
                      {-fol-plugin( monadM.proofs.law3 A {A;S} {B;S}
                                  , extensionality
                                  )-}
          )
      )

-- The State monad --------------------------------------------------------

State : Set -> Set -> Set
State S = StateT S Id

monadState : (S : Set) -> Monad (State S)
monadState S = transformMonad (StateT S)  (monadTransStateT S)
                              Id          monadId

-- State monad operations -------------------------------------------------

