{-# OPTIONS -fglasgow-exts #-}

{-| Names (identifiers) and creating of fresh names.

A name is a string which is composed of a stem and optionally a
version (decimal number).  For example, the identifier "bla3" is
interpreted as stem "bla", version 3.  If the version is missing, it
is interpreted as "-1".  Equality on identifiers is just the literal
String equality; stem and version are only relevant for the creation
of fresh names.  To this end, a data structure 'UsedNames',
implemented as a finite map, keeps track of the next free identifier
version for each stem.  If a fresh name for a certain stem is needed,
the highest version is retrieved from UsedNames and UsedNames is
incremented for this stem.

-}
 
module Utils.Names
 ( UsedNames
 , noNamesUsed
 -- , stem
 , quoted
 , fresh --, localFresh
 , HasNames(..)
 ) where

import Control.Monad.State
import Control.Monad.Reader
import Data.Char	    ( isDigit )

import Syntax.Position
import Syntax.Name
import Utils.Map
import Utils.Monads

type VName = (Int, Name)

-- Note: We need overlapping instances for this.
-- And it's not strictly neccessary.
{-instance Show VName where
    show (i, x) = show x-}

-- | stem drops the trailing digits in a name
-- stem :: Name -> Name
-- stem = quoted (reverse . dropWhile isDigit . reverse)

quoted f ('"':x)    = "\"" ++ f (init x) ++ "\""
quoted f x	    = f x

type UsedNames = Int

noNamesUsed :: UsedNames
noNamesUsed = 0

-- | registers the next unused version of a name
freshName :: String -> UsedNames -> (UsedNames, Name)
--freshName "_" uid   = (uid, DontCare) we wan't fresh underscores as well
freshName x uid	    = (uid + 1, Name noRange x uid)

class HasNames n where
    getNames :: n -> UsedNames
    setNames :: UsedNames -> n -> n

fresh :: (HasNames n, MonadState n m) => String -> m Name
fresh x =
    do	(used',x') <- freshName x <$> (getNames <$> get)
	modify (setNames used')
	return x'

instance HasNames UsedNames where
    getNames = id
    setNames = const

